/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>> 	Main program for the edge program - with required GDAL dependencies
   >>>>
   >>>>  Private:
   >>>> 	main(int argc, char **argv)
   >>>> 	usage()
   >>>> 	help()
   >>>>
   >>>>   Static:
   >>>>   Public:
   >>>>
   >>>>       Written: April 15, 2013
   >>>> Modifications: April 22, 2013 - Added edge_threshold parameter and changed edge function call interface.
   >>>>                May   22, 2013 - Added scale_output and output_mask_image parameters
   >>>>                May   28, 2013 - Replaced band_average flag with output_type parameter
   >>>>                July  26, 2013 - Added "Std_Dev" operation type and optional input mask 
   >>>>                Sept. 25, 2013 - Added the Frei-Chen_option parameter for edge, line and edge-line options of Frei-Chen operator
   >>>>                April 19, 2014 - Modified to require GDAL and call GDAL utilities directly
   >>>>                April 19, 2014 - Removed "Std_Dev" option
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "edge.h"
#include "params/params.h"
#include <gdal_priv.h>
#include <iostream>
#include <cstdlib>
#include <climits>
#include <cstring>

using namespace std;
using namespace CommonTilton;

//Globals
Params params("Version 1.70G, April 19, 2016");

// Forward function declarations
void usage();
void help();

/*-----------------------------------------------------------
|
|  Routine Name: main - Main program for the edge program interface
|
|  Purpose: Declares program parameters, reads parameter file and initializes program parameters,
|           opens log file (if requested), calls the select edge function (Prewitt, Sobel, Scharr or Frei-Chen) 
|           and returns an exit status.
|
|  Input from command line - parameter_file_name
|
|       Returns: EXIT_SUCCESS on success, EXIT_FAILURE on failure
|
|    Written By: James C. Tilton, NASA's GSFC, Mail Code 606.3, Greenbelt,MD 20771
|        E-Mail: James.C.Tilton@nasa.gov
|
|       Written: April 15, 2014
| Modifications: August 12, 2014 - Bug fixes.
|
------------------------------------------------------------*/
int main(int argc, char *argv[])
{
  bool status = true;

  GDALAllRegister();

  if (argc == 1)
  {
    usage();
    cout << "ERROR: Need parameter file as argument." << endl;
    return EXIT_FAILURE;
  }
  else if ((strncmp(argv[1],"-v",2) == 0) || (strncmp(argv[1],"-version",8) == 0))
  {
    params.print_version();
    cout << "For help information: edge -h or edge -help" << endl << endl;
    return EXIT_SUCCESS;
  }
  else if ((strncmp(argv[1],"-h",2) == 0) || (strncmp(argv[1],"-help",5) == 0))
  {
    help();
    return EXIT_SUCCESS;
  }
  else if (strncmp(argv[1],"-",1) == 0)
  {
    usage();
    cout << "ERROR: The parameter file name cannot start with an \"-\"" << endl;
    return EXIT_FAILURE;
  }
  else
  {
    if (argc != 2)
    {
      usage();
      cout << "ERROR: Incorrect number of parameters on command line" << endl;
      return EXIT_FAILURE;
    }
    else
    {
      status = params.read(argv[1]);
      if (!status)
      {
        usage();
        cout << "ERROR: Error reading parameter file (read)" << endl;
        return EXIT_FAILURE;
      }
    }
  }

  if (status)
  {
    params.print();
    status = edge();
  }

  if (status)
  {
    cout << endl << "Successful completion of edge program" << endl;
  }
  else
  {
    cout << endl << "The edge program terminated improperly." << endl;
  }
  
  if (status)
    return EXIT_SUCCESS;
  else
    return EXIT_FAILURE;
}

/*-----------------------------------------------------------
|
|  Routine Name: usage - Usage function
|
|       Purpose: Informs user of proper usage of program when mis-used.
|
|         Input:
|
|        Output:
|
|       Returns: (void)
|
|    Written By: James C. Tilton, NASA's GSFC, Mail Code 606.3, Greenbelt,MD 20771
|        E-Mail: James.C.Tilton@nasa.gov
|
|       Written: April 15, 2013
| Modifications: 
|
------------------------------------------------------------*/
void usage()
{
  cout << endl << "Usage: " << endl << endl;
  cout << "edge parameter_file_name" << endl << endl;
  cout << "For help information: edge -h or edge -help" << endl;
  cout << "For version information: edge -v or edge -version" << endl;

  return;
}

/*-----------------------------------------------------------
|
|  Routine Name: help - Help function
|
|       Purpose: Provides help information to user on program parameters
|
|         Input:
|
|        Output:
|
|       Returns: (void)
|
|    Written By: James C. Tilton, NASA's GSFC, Mail Code 606.3, Greenbelt,MD 20771
|        E-Mail: James.C.Tilton@nasa.gov
|
|       Written: April 15, 2013
| Modifications: April 22, 2013 - Added edge_threshold parameter
|                May   22, 2013 - Added scale_output and output_mask_image parameters
|
------------------------------------------------------------*/
void help()
{
  cout << endl << "The edge progam is called in the following manner:" << endl;
  cout << endl << "edge parameter_file_name" << endl;
  cout << endl << "where \"parameter_file_name\" is the name of the input parameter" << endl;
  cout << "file. For contents see below." << endl;
  cout << endl << "For this help: edge -h or edge -help" << endl;
  cout << endl << "For version information: edge -v or edge -version";
  cout << endl;
  cout << endl << "The parameter file consists of entries of the form:" << endl;
  cout << endl << "-parameter_name parameter_value(s)" << endl;

  fprintf(stdout,"\nThe following parameters may be specified in the input parameter file:\n\n");
  fprintf(stdout,"Input File and parameter:\n"
"-input_image		(string)	Input image file name(required)\n"
"					Integer data or float data types of\n"
"					up through 32 bits are allowed\n"
"-mask			(string)	Input image mask file name\n"
"					Data type must be \'Byte\'.\n"
"					NOTE: The input image may be instead\n"
"					masked by specifying a \'NoData Value\'\n"
"					in the input image.\n"
"					(optional, default = {none})\n"
"-mask_value		(int)		If input mask file is provided,\n"
"					this is the value in the mask file that\n"
"					designates invalid data. (default is\n"
"					 \'NoData Value\' from image header,\n"
"					 if provided - otherwise 0.)\n"
"-bias_value		(int)		Bias value to add to the input data\n"
"					values before the edge value is\n"
"					computed. (Optional, default = 0.)\n"
"-edge_operation		(string)	Edge operation type (required)\n"
"					Must be either:\n"
"					  Prewitt\n"
"					  Sobel\n"
"					  Scharr or\n"
"					  Frei-Chen\n"
"-Frei-Chen_option		(int)	Frei-Chen operation option:\n"
"					  1: edge\n"
"					  2: line-edge\n"
"					(default, 1: edge. Valid only for\n"
"					 the Frei-Chen edge_operation.)\n"
"-edge_threshold		(float)		Edge threshold value (default = 0.0)\n"
"-output_type		(int)		Output Type:\n"
"					  1. \"Multispectral,\"\n"
"					  2. \"Band Average,\"\n"
"					  3. \"Band Maximum,\"\n"
"					  4. \"Band Minimum,\"\n"
"					(default = 1 -> \"Multispectral\")\n"
"-scale_output		(bool)		Flag to request scaling of output edge\n"
"					values to std. dev. = 1.0 and min. = 0.0.\n"
"					(true (1) or false (0), default = false)\n");
  fprintf(stdout,"Output File:\n"
"-output_image		(string)	Output image (required)\n"
"-output_mask_image	(string)	Output mask image (required)\n");

  return;
}

